package org.openxma.dsl.generator.impl;

import static org.openxma.dsl.common.extensions.EObjectPropertiesAdapter.setProperty;

import java.util.List;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.xtext.naming.DefaultDeclarativeQualifiedNameProvider;
import org.eclipse.xtext.naming.IQualifiedNameProvider;
import org.eclipse.xtext.naming.QualifiedName;
import org.openxma.dsl.core.CoreFactory;
import org.openxma.dsl.core.model.DataTypeAndTypeParameter;
import org.openxma.dsl.core.model.EnumType;
import org.openxma.dsl.core.model.ParameterDefinition;
import org.openxma.dsl.core.model.ParameterDefinitionType;
import org.openxma.dsl.core.model.SimpleType;
import org.openxma.dsl.core.model.SqlType;
import org.openxma.dsl.core.model.StringParameterValue;
import org.openxma.dsl.core.model.Type;
import org.openxma.dsl.dom.DomFactory;
import org.openxma.dsl.dom.model.Attribute;
import org.openxma.dsl.dom.model.AttributeGroup;
import org.openxma.dsl.dom.model.AttributeSortOrder;
import org.openxma.dsl.dom.model.CallOutputParameter;
import org.openxma.dsl.dom.model.CallableStatement;
import org.openxma.dsl.dom.model.Column;
import org.openxma.dsl.dom.model.ComplexType;
import org.openxma.dsl.dom.model.Dao;
import org.openxma.dsl.dom.model.DataBaseConstraint;
import org.openxma.dsl.dom.model.DataBaseConstraintType;
import org.openxma.dsl.dom.model.Entity;
import org.openxma.dsl.dom.model.ManyToMany;
import org.openxma.dsl.dom.model.ManyToOne;
import org.openxma.dsl.dom.model.OneToMany;
import org.openxma.dsl.dom.model.OneToOne;
import org.openxma.dsl.dom.model.QueryOperation;
import org.openxma.dsl.dom.model.ValueObject;
import org.openxma.dsl.generator.GeneratorConfiguration;
import org.openxma.dsl.generator.ModelModifier;

import com.google.common.base.Function;
import com.google.common.base.Predicate;
import com.google.common.collect.Collections2;
import com.google.common.collect.ConcurrentMultiset;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Iterables;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;

/**
 * Represents a {@link ModelModifier} used to derive and create a default
 * database mapping from a given given {@link Dao Repository}.
 */
public class DefaultRepositoryModelModifier extends AbstractDomModelModifier<Void> {
	/**
	 * hibernate related constans for userTypes and sqlTypes
	 */
	public static final String HIBERNATE_USERTYPE_PROPERTY_POSTFIX = "userType";
	public static final String HIBERNATE_SQLTYPE_PROPERTY_POSTFIX = "sqlType";
	public static final String HIBERNATE_USERTYPE_PROPERTY_FORMAT = "hibernate.%s." + HIBERNATE_USERTYPE_PROPERTY_POSTFIX;
	public static final String HIBERNATE_SQLTYPE_PROPERTY_FORMAT = "hibernate.%s." + HIBERNATE_SQLTYPE_PROPERTY_POSTFIX;

	@Override
	public Void caseDao(Dao dao) {
		reconcileId(dao);
		deriveTableInformations(dao);
		deriveColumns(dao);
		deriveRelations(dao);
		deriveConstraints(dao);
		augmentConstraints(dao);
		augmentCallableStatements(dao);
		return super.caseDao(dao);
	}

	
    private void augmentCallableStatements(Dao dao) {
        for (QueryOperation queryOperation : dao.getQueryOperation()) {
            if (queryOperation.getStatement() instanceof CallableStatement
                    && queryOperation.getType() instanceof ComplexType) {
                CallableStatement callableStatement = (CallableStatement) queryOperation.getStatement();
                ComplexType complexType = (ComplexType) queryOperation.getType();
                ImmutableMap<String, Attribute> attributeMap = Maps.uniqueIndex(complexType.getAllAttributes(),
                        new Function<Attribute, String>() {
                            public String apply(Attribute from) {
                                return from.getName();
                            }
                        });
                for (CallOutputParameter callOutputParameter : callableStatement.getOutParameter()) { 
                    if (callOutputParameter.getAttribute() == null) {
                        callOutputParameter.setAttribute(attributeMap.get(callOutputParameter.getName()));
                    }
                }
            }
        }
    }

    private void reconcileId(Dao dao) {
		Entity entity = dao.getEntity();
		AttributeGroup naturalKeyGroup = entity.getKey();
		Attribute surrogateIdentifier = (Attribute) entity.getIdentifier();
		if (naturalKeyGroup != null) {
			DataBaseConstraint naturalKey = DomFactory.eINSTANCE.createDataBaseConstraint();
			naturalKey.setType(DataBaseConstraintType.NATURAL);
			naturalKey.setName(naturalKeyGroup.getName());
			naturalKey.getAttributes().addAll(
					Collections2.forIterable(Iterables.transform(naturalKeyGroup.getAttributes(),
							new Function<AttributeSortOrder, Attribute>() {
								public Attribute apply(AttributeSortOrder from) {
									return from.getAttribute();
								}
							})));
			dao.getDataBaseConstraints().add(naturalKey);
		}
		if (surrogateIdentifier != null) {
			DataBaseConstraint primaryKey = DomFactory.eINSTANCE.createDataBaseConstraint();
			primaryKey.setType(DataBaseConstraintType.PRIMARY);
			primaryKey.setName(entity.getName() + "PrimaryKey");
			primaryKey.getAttributes().add(surrogateIdentifier);
			dao.getDataBaseConstraints().add(primaryKey);
		}
	}

	private void deriveTableInformations(Dao dao) {
		if (dao.getTableName() == null) {
			dao.setTableName(GeneratorConfiguration.INSTANCE.getNamingStrategy().getTableName(dao));
		}
		if (dao.getQualifier() == null) {
			dao.setQualifier(GeneratorConfiguration.INSTANCE.getNamingStrategy().getQualifier(dao));
		}
		if (dao.getDiscriminator() == null) {
			dao.setDiscriminator("DISCRIMINATOR");
		}
	}

	private void deriveColumns(Dao dao) {
		for (Attribute attribute : dao.getEntity().copyAttributes()) {
			if (!attribute.isTransient() && !attribute.isReference()) {
				Column column = dao.getColumnDefinition(attribute);
				if (null == column) {
					column = DomFactory.eINSTANCE.createColumn();
					column.setAttribute(attribute);
					dao.getColumns().add(column);
					if (attribute.getType().getDataType() instanceof ValueObject) {
						addComponentTypeColumnMapping(dao.getEntity(), attribute, column);
					} else if (isValueObjectType(attribute)) {
						setValueObjectType(attribute, column);
					}
				} else if (attribute.getType().getDataType() instanceof ValueObject) {
                    ComplexType complexType = (ComplexType) attribute.getType().getDataType();
                    EList<Attribute> attributes = complexType.getAttributes();
                    for (Attribute nestedAttribute : attributes) {
                        Column nestedColumn = column.getColumnDefinition(nestedAttribute);
                        if (nestedColumn == null) {
                            nestedColumn = DomFactory.eINSTANCE.createColumn();
                            nestedColumn.setAttribute(nestedAttribute);
                            column.getColumns().add(nestedColumn);
                        }
                        if (nestedColumn.getColumnName() == null) {
                            nestedColumn.setColumnName(GeneratorConfiguration.INSTANCE.getNamingStrategy().getColumnName(
                                    dao.getEntity(), attribute, nestedAttribute));
                        }
                        applyColumnUserType(nestedColumn);
                        applyColumnSqlType(nestedColumn);
                    }
				}
                if (column.getColumnName() == null) {
                    column.setColumnName(GeneratorConfiguration.INSTANCE.getNamingStrategy().getColumnName(dao.getEntity(),
                            attribute));
                }
				applyColumnUserType(column);
				applyColumnSqlType(column);
			}
		}
	}

	private boolean isValueObjectType(Attribute attribute) {
		return attribute.getDataType() instanceof SimpleType && ((SimpleType) attribute.getDataType()).isCustom();
	}

	private void applyColumnSqlType(Column column) {
		Type type = column.getAttribute().getDataType();
		if (column.getColumnType() == null && type instanceof SimpleType) {
			SimpleType simpleType = (SimpleType) type;
			String javaType = new QualifiedName(simpleType.getInstanceTypeName()).lastSegment();
			String sqlType = GeneratorConfiguration.INSTANCE.getProperty(String.format(HIBERNATE_SQLTYPE_PROPERTY_FORMAT, javaType));
			if (sqlType != null) {
				column.setColumnType(createSqlType(javaType + "SqlType", sqlType));
			}
		}
	}

	private void applyColumnUserType(Column column) {
		Type type = column.getAttribute().getDataType();
		if (column.getUserType() == null && type instanceof SimpleType) {
			SimpleType simpleType = (SimpleType) type;
			if (type instanceof EnumType) {
				setEnumUserType(column, type);
			} else {
				setUserTypeFromProperties(column, simpleType);
			}
		}
	}

	/**
	 * @deprecated ValueObjects are deprecated
	 */
	private void setValueObjectType(Attribute attribute, Column column) {
		SimpleType simple = (SimpleType) attribute.getType().getDataType();
		SimpleType simpleType = CoreFactory.eINSTANCE.createSimpleType("ValueObjectHibernateType",
				"org.openxma.dsl.platform.hibernate.type.ValueObjectHibernateType");
		ParameterDefinition componentClassParameter = CoreFactory.eINSTANCE.createParameterDefinition();
		componentClassParameter.setName("valueObjectClass");
		componentClassParameter.setType(ParameterDefinitionType.STRING);
		simpleType.getTypeParameter().add(componentClassParameter);
		DataTypeAndTypeParameter userType = createDataTypeAndTypeParameter(simpleType);
		StringParameterValue componentClass = CoreFactory.eINSTANCE.createStringParameterValue();
		String fullyQualifiedName = getFullyQualifiedName(simple);
		componentClass.setStringValue(fullyQualifiedName);
		userType.getParameterValues().add(componentClass);
		column.setUserType(userType);
	}

	private void setUserTypeFromProperties(Column column, SimpleType simpleType) {
		String javaType = new QualifiedName(simpleType.getInstanceTypeName()).lastSegment();
		String userTypeClass = GeneratorConfiguration.INSTANCE.getProperty(String.format(HIBERNATE_USERTYPE_PROPERTY_FORMAT, javaType));
		if (userTypeClass != null) {
			column.setUserType(createDataTypeAndTypeParameter(CoreFactory.eINSTANCE.createSimpleType(javaType + "UserType", userTypeClass)));
		}
	}

	private void setEnumUserType(Column column, Type type) {
		String userTypeClass = "org.openxma.dsl.platform.hibernate.type.GenericEnumUserType";
		SimpleType enumType = CoreFactory.eINSTANCE.createSimpleType(column.getAttribute().getName() + "UserType", userTypeClass);
		ParameterDefinition componentClassParameter = CoreFactory.eINSTANCE.createParameterDefinition();
		componentClassParameter.setName("enumClass");
		componentClassParameter.setType(ParameterDefinitionType.STRING);
		enumType.getTypeParameter().add(componentClassParameter);
		DataTypeAndTypeParameter userType = createDataTypeAndTypeParameter(enumType);
		StringParameterValue enumClass = CoreFactory.eINSTANCE.createStringParameterValue();
		String fullyQualifiedName = getFullyQualifiedName((EnumType) type);
		enumClass.setStringValue(fullyQualifiedName);
		userType.getParameterValues().add(enumClass);
		column.setUserType(userType);
	}

	private void deriveRelations(Dao dao) {
		deriveToOneRelations(dao);
		deriveToManyReferences(dao);
	}

	private void deriveToManyReferences(Dao dao) {
		Iterable<Attribute> toManyReferences = Iterables.filter(dao.getEntity().getAttributes(), new Predicate<Attribute>() {
			public boolean apply(Attribute reference) {
				return reference.isReference() && reference.isMany() && !reference.isTransient();
			}
		});
		for (Attribute reference : toManyReferences) {
			if (reference.getOpposite() != null) {
				Entity oppositeEntity = (Entity) reference.getOpposite().eContainer();
				deriveToOneRelations(oppositeEntity.getRepository());
			}
		}
		deriveRelations(dao, toManyReferences);
	}

	private void deriveToOneRelations(Dao dao) {
		if (dao != null) {
			Iterable<Attribute> toOneReferences = Iterables.filter(dao.getEntity().getAttributes(),
					new Predicate<Attribute>() {
						public boolean apply(Attribute reference) {
							return reference.isReference() && !reference.isMany() && !reference.isTransient();
						}
					});
			deriveRelations(dao, toOneReferences);
		}
	}

	private void deriveRelations(Dao dao, Iterable<Attribute> toOneReferences) {
		Entity entity = dao.getEntity();
		for (Attribute reference : toOneReferences) {
			if (reference.isMany()) {
				deriveCollectionMapping(dao, entity, reference);
			} else {
				deriveEntityMapping(dao, entity, reference);
			}
		}
	}

	private void deriveEntityMapping(Dao dao, Entity entity, Attribute attribute) {
		if (attribute.getOpposite() == null) {
			ManyToOne manyToOne = dao.getManyToOne(attribute);
			if (manyToOne == null) {
				manyToOne = DomFactory.eINSTANCE.createManyToOne();
				manyToOne.setAttribute(attribute);
				dao.getManyToOneAssociations().add(manyToOne);
			}
			Entity referencedEntity = (Entity) manyToOne.getAttribute().getDataType();
			if (referencedEntity.getIdentifier() != null) {
				if (manyToOne.getColumnName() == null) {
					manyToOne.setColumnName(GeneratorConfiguration.INSTANCE.getNamingStrategy().getColumnName(entity, attribute));
				}
			} else if (referencedEntity.getKey() != null) {
				if (manyToOne.getColumns().isEmpty()) {
					AttributeGroup key = referencedEntity.getKey();
					for (Attribute attributeHolder : key.getResolvedAttributeList()) {
						Column column = DomFactory.eINSTANCE.createColumn();
						column.setAttribute(attributeHolder);
						manyToOne.getColumns().add(column);
					}
				}
				for (Column column : manyToOne.getColumns()) {
					if (column.getColumnName() == null) {
						column.setColumnName(GeneratorConfiguration.INSTANCE.getNamingStrategy().getColumnName(entity,
								column.getAttribute()));
					}
				}
			}
		} else {
			OneToOne oneToOne = dao.getOneToOne(attribute);
			if (null == oneToOne) {
				oneToOne = DomFactory.eINSTANCE.createOneToOne();
				oneToOne.setAttribute(attribute);
				dao.getOneToOneAssociations().add(oneToOne);
			}
		}
	}

	private void deriveCollectionMapping(Dao dao, Entity entity, Attribute reference) {
		ManyToMany manyToMany = dao.getManyToMany(reference);
		if (null != manyToMany) {
			if (manyToMany.getColumnName() == null) {
				Column identifierColumn = dao.getColumnDefinition((Attribute) entity.getIdentifier());
				manyToMany.setColumnName(identifierColumn.getColumnName());
			}
			if (reference.getOpposite() != null) {
				Attribute opposite = reference.getOpposite();
				if (null == opposite.getOpposite()) {
					manyToMany.setInverse(true);
				} else {
					// mark any side as inverse=true doesn't matter which side
					IQualifiedNameProvider nameProvider = new DefaultDeclarativeQualifiedNameProvider();
					String attributeName = nameProvider.getQualifiedName(reference);
					String oppositeAttributeName = nameProvider.getQualifiedName(opposite);
					Boolean alreadySetInverseOnOtherSide = false;
					if (attributeName.length() == oppositeAttributeName.length()) {
						Entity referencedEntity = (Entity) reference.getDataType();
						Dao repository = referencedEntity.getRepository();
						ManyToMany oppositeManyToMany = repository.getManyToMany(opposite);
						alreadySetInverseOnOtherSide = (oppositeManyToMany != null && oppositeManyToMany.isInverse());
					}
					manyToMany.setInverse(attributeName.length() <= oppositeAttributeName.length()
							&& !alreadySetInverseOnOtherSide);
				}
			}
		} else {
			OneToMany oneToMany = dao.getOneToMany(reference);
			if (null == oneToMany) {
				oneToMany = DomFactory.eINSTANCE.createOneToMany();
				oneToMany.setAttribute(reference);
				dao.getOneToManyAssociations().add(oneToMany);
			}
			if (entity.getIdentifier() != null) {
				if (oneToMany.getColumnName() == null) {
					String columnName = GeneratorConfiguration.INSTANCE.getNamingStrategy().getColumnName(entity, reference);
					if (reference.getOpposite() != null) {
						Entity oppositeEntity = (Entity) reference.getOpposite().eContainer();
						ManyToOne manyToOne = oppositeEntity.getRepository() != null ? oppositeEntity.getRepository()
								.getManyToOne(reference.getOpposite()) : null;
						if (null != manyToOne) {
							columnName = manyToOne.getColumnName();
						}
					}
					oneToMany.setColumnName(columnName);
				}
			} else if (entity.getKey() != null) {
				if (oneToMany.getColumns().isEmpty()) {
					if (reference.getOpposite() != null) {
						Entity oppositeEntity = (Entity) reference.getOpposite().eContainer();
						ManyToOne manyToOne = oppositeEntity.getRepository().getManyToOne(reference.getOpposite());
						oneToMany.getColumns().addAll(EcoreUtil.copyAll(manyToOne.getColumns()));
					} else {
						for (Attribute attributeHolder : entity.getKey().getResolvedAttributeList()) {
							Column column = DomFactory.eINSTANCE.createColumn();
							column.setAttribute((Attribute) attributeHolder);
							oneToMany.getColumns().add(column);
						}
					}
				}
				for (Column column : oneToMany.getColumns()) {
					if (column.getColumnName() == null) {
						column.setColumnName(GeneratorConfiguration.INSTANCE.getNamingStrategy().getColumnName(entity,
								column.getAttribute()));
					}
				}
			}
		}
	}

	private void deriveConstraints(Dao dao) {
		Entity entity = dao.getEntity();
		for (AttributeGroup featureGroup : entity.getAttributeGroups()) {
			if (featureGroup.isUnique() || featureGroup.isFilter()) {
				DataBaseConstraint uniqueConstraint = DomFactory.eINSTANCE.createDataBaseConstraint();
				uniqueConstraint.setType(featureGroup.isUnique() ? DataBaseConstraintType.UNIQUE
						: DataBaseConstraintType.INDEX);
				uniqueConstraint.setName(featureGroup.getName());
				uniqueConstraint.getAttributes().addAll(
						Collections2.forIterable(Iterables.transform(featureGroup.getAttributes(),
								new Function<AttributeSortOrder, Attribute>() {
									public Attribute apply(AttributeSortOrder from) {
										return from.getAttribute();
									}
								})));
				dao.getDataBaseConstraints().add(uniqueConstraint);
			}
		}
	}

	private void augmentConstraints(Dao dao) {
		ConcurrentMultiset<DataBaseConstraint> uniqueKeys = new ConcurrentMultiset<DataBaseConstraint>(dao
				.getDataBaseConstraintByType(DataBaseConstraintType.UNIQUE));
		for (DataBaseConstraint dataBaseConstraint : uniqueKeys) {
			EList<Attribute> fields = dataBaseConstraint.getAttributes();
			boolean containsReferences = Iterables.any(fields, new Predicate<Attribute>() {
				public boolean apply(Attribute structuralFeature) {
					return structuralFeature.isReference();
				}
			});
			if (containsReferences) {
				DataBaseConstraint resolvedDataBaseConstraint = DomFactory.eINSTANCE.createDataBaseConstraint();
				resolvedDataBaseConstraint.setName(dataBaseConstraint.getName());
				resolvedDataBaseConstraint.setType(dataBaseConstraint.getType());
				List<String> references = Lists.newArrayList();
				for (Attribute dataBaseField : fields) {
					if (dataBaseField.isReference()) {
						references.add(dataBaseField.getName());
						Entity referencedEntity = (Entity) dataBaseField.getDataType();
						if (referencedEntity.getIdentifier() != null) {
							Attribute identifierAttribute = (Attribute) EcoreUtil.copy(referencedEntity.getIdentifier());
							setProperty(identifierAttribute, "fqn", referencedEntity.getName());
							resolvedDataBaseConstraint.getAttributes().add(identifierAttribute);
						} else if (referencedEntity.getKey() != null) {
							for (AttributeSortOrder structuralFeature : referencedEntity.getKey().getAttributes()) {
								Attribute naturalKeyFeature = (Attribute) EcoreUtil.copy(structuralFeature.getAttribute());
								setProperty(naturalKeyFeature, "fqn", referencedEntity.getName());
								resolvedDataBaseConstraint.getAttributes().add(naturalKeyFeature);
							}
						}
					} else {
						resolvedDataBaseConstraint.getAttributes().add((Attribute) EcoreUtil.copy(dataBaseField));
					}
				}
				setProperty(resolvedDataBaseConstraint, "alias", references);
				dao.getDataBaseConstraints().add(resolvedDataBaseConstraint);
			}
		}
	}

	private DataTypeAndTypeParameter createDataTypeAndTypeParameter(SimpleType simpleType) {
		DataTypeAndTypeParameter dataTypeAndTypeParameter = CoreFactory.eINSTANCE.createDataTypeAndTypeParameter();
		dataTypeAndTypeParameter.setDataType(simpleType);
		return dataTypeAndTypeParameter;
	}

	

	private SqlType createSqlType(String name, String ddl) {
		SqlType sqlType = CoreFactory.eINSTANCE.createSqlType();
		sqlType.setName(name);
		sqlType.setDdl(ddl);
		return sqlType;
	}

	private void addComponentTypeColumnMapping(Entity entity, Attribute attribute, Column column) {
		ComplexType complexType = (ComplexType) attribute.getType().getDataType();
		EList<Attribute> attributes = complexType.getAttributes();
		for (Attribute nestedAttribute : attributes) {
			Column nestedColumn = DomFactory.eINSTANCE.createColumn();
			nestedColumn.setAttribute(nestedAttribute);
			column.getColumns().add(nestedColumn);
			String columnName = GeneratorConfiguration.INSTANCE.getNamingStrategy().getColumnName(entity, attribute, nestedAttribute);
			nestedColumn.setColumnName(columnName);
		}
	}

	private String getFullyQualifiedName(SimpleType simpleType) {
		if (isGeneratedSimpleType(simpleType)) {
			return GeneratorConfiguration.INSTANCE.getNamingStrategy().getPackageName(simpleType) + "." + simpleType.getName();
		}
		return simpleType.getInstanceTypeName();
	}

	private boolean isGeneratedSimpleType(SimpleType simpleType) {
		return isGeneratedValueObject(simpleType) || simpleType instanceof EnumType;
	}

	private boolean isGeneratedValueObject(SimpleType simpleType) {
		return simpleType.isCustom() && simpleType.getTypeDefinition().getDefinition() != null;
	}
}
